;;; Metamail interface for GNU Emacs
;; Copyright (C) 1993 Masanobu UMEDA
;; $Header: metamail.el,v 1.1 93/03/02 00:51:50 umerin Locked $

;; This file is part of GNU Emacs.

;; GNU Emacs is distributed in the hope that it will be useful,
;; but WITHOUT ANY WARRANTY.  No author or distributor
;; accepts responsibility to anyone for the consequences of using it
;; or for whether it serves any particular purpose or works at all,
;; unless he says so in writing.  Refer to the GNU Emacs General Public
;; License for full details.

;; Everyone is granted permission to copy, modify and redistribute
;; GNU Emacs, but only under the conditions described in the
;; GNU Emacs General Public License.   A copy of this license is
;; supposed to have been given to you along with GNU Emacs so you
;; can know your rights and responsibilities.  It should be in a
;; file named COPYING.  Among other things, the copyright notice
;; and this notice must be preserved on all copies.

;; LCD Archive Entry:
;; metamail|Masanobu UMEDA|umerin@mse.kyutech.ac.jp|
;; Metamail interface for GNU Emacs|
;; 93-02-20|1.0|~/misc/metamail.el.Z|

;; Basic ideas from gnus-mime.el by Spike <Spike@world.std.com>
;; with code from Michael Littman's <mlittman@breeze.bellcore.com>
;; richtext.el and metamail's MH-E patches.

;; Note: Metamail does not have all options which is compatible with
;; the environment variables.  For that reason, matamail.el have to
;; hack the environment variables.  In addition, there is no way to
;; display all header fields without extra informative body messages
;; which is suppressed by "-q" option.

(provide 'metamail)

(defvar metamail-program-name "metamail"
  "*Metamail program name.")

(defvar metamail-environment "KEYHEADS='*';export KEYHEADS;"
  "*Environment variables for Metamail.
It must be an emtpy string or a string terminated with ';'.")

(defvar metamail-switches '("-m" "emacs" "-x" "-d" "-z")
  "*Switches for Metamail program.
-z is required to remove zap file.")

(defun metamail-buffer (&optional buffer)
  "Process current buffer through 'metamail'.
Optional argument BUFFER specifies a buffer to be filled (nil means current)."
  (interactive)
  (metamail-region (point-min) (point-max) buffer))

(defun metamail-region (beg end &optional buffer)
  "Process current region through 'metamail'.
Optional argument BUFFER specifies a buffer to be filled (nil means current)."
  (interactive "r")
  (let ((curbuf (current-buffer))
	(buffer-read-only nil)
	(metafile (make-temp-name "/tmp/metamail")))
    (save-excursion
      ;; Gee!  Metamail does not ouput to stdout if input comes from
      ;; stdin.
      (write-region (point-min) (point-max) metafile nil 'nomessage)
      (if (eq curbuf
	      (if buffer (get-buffer buffer) (current-buffer)))
	  (delete-region (point-min) (point-max)))
      ;; We have to pass the environment variable KEYHEADS to /bin/sh
      ;; to display all header fields.  Metamail should have an
      ;; optional argument to pass such information directly.
      (apply (function call-process)
	     "/bin/sh"
	     nil
	     (or buffer t)		;BUFFER or current buffer
	     t				;Force redisplay
	     (list "-c"
		   ;; Construct environment and the command.
		   (concat
		    metamail-environment
		    metamail-program-name
		    " "
		    (mapconcat (function identity) metamail-switches " ")
		    " "
		    metafile
		    )))
      )))

;(defun metamail-region (beg end &optional buffer)
;  "Process current region through 'metamail'.
;Optional argument BUFFER specifies a buffer to be filled (nil means current)."
;  (interactive "r")
;  (let ((curbuf (current-buffer))
;	(buffer-read-only nil)
;	(metafile (make-temp-name "/tmp/metamail")))
;    (save-excursion
;      (write-region (point-min) (point-max) metafile nil 'nomessage)
;      (if (eq curbuf
;	      (if buffer (get-buffer buffer) (current-buffer)))
;	  (delete-region (point-min) (point-max)))
;      (apply (function call-process)
;	     metamail-program-name
;	     nil
;	     (or buffer t)		;BUFFER or current buffer
;	     nil			;don't redisplay
;	     (append metamail-switches (list metafile)))
;      )))
